
'''
This file is automatically generated; Do not edit it. :)
'''
VERSION_INFO = {
    'final': True,
    'version': '1.26.9',
    'branch_nick': 'release-1.26-LTS',
    'revision_id': '67b70d9d1fdf3a9a0eaa17205c5705276ba64c60',
    'revno': 10917
}
